@@ -9,12 +9,13 @@ use strict;
 use warnings;
 use 5.010;
 
-our $VERSION = '0.11';
+our $VERSION = '0.12';
 
 use base qw( IO::Async::Notifier );
 
 use Carp;
 
+use Devel::Refcount qw/ refcount /;
 use Future::Utils qw( fmap_void try_repeat_until_success );
 use List::Util qw( shuffle );
 use Scalar::Util qw( weaken );
@@ -29,6 +30,11 @@ use constant DEFAULT_CQL_PORT => 9042;
 # Time after which down nodes will be retried
 use constant NODE_RETRY_TIME => 60;
 
+# How long after a query is last used to keep it alive on a pacemaker
+use constant QUERY_TTL => 60;
+# How often the pacemaker checks the query isn't the only thing keeping the whole stack alive
+use constant QUERY_PACEMAKER_INTERVAL => 0.5;
+
 =head1 NAME
 
 C<Net::Async::CassandraCQL> - use Cassandra databases with L<IO::Async> using CQL
@@ -43,7 +49,7 @@ C<Net::Async::CassandraCQL> - use Cassandra databases with L<IO::Async> using CQ
 
  my $cass = Net::Async::CassandraCQL->new(
     host => "localhost",
-    keyspace => "my-keyspace",
+    keyspace => "my_keyspace",
     default_consistency => CONSISTENCY_QUORUM,
  );
  $loop->add( $cass );
@@ -54,7 +60,7 @@ C<Net::Async::CassandraCQL> - use Cassandra databases with L<IO::Async> using CQ
 
  my @f;
  foreach my $number ( 1 .. 100 ) {
-    push @f, $cass->query( "INSERT INTO numbers (v) VALUES $number" );
+    push @f, $cass->query( "INSERT INTO numbers (v) VALUES ($number)" );
  }
  Future->needs_all( @f )->get;
 
@@ -119,6 +125,8 @@ The hostnames of Cassandra node to connect to initially. If more than one host
 is provided in an array, they will be attempted sequentially until one
 succeeds during the intial connect phase.
 
+If a host contains :$service, then the service parameter will be overidden for this host.
+
 =item service => STRING
 
 Optional. The service name or port number to connect to.
@@ -173,7 +181,7 @@ sub _init
       shift->_on_status_change( @_ );
    });
 
-   $self->{queries_by_cql} = {}; # {$cql} => [$query, $expire_timer_f]
+   $self->{queries_by_cql} = {}; # {$cql} => {query => $query, pacemaker => $expire_timer_f, ttl => Int]
    # Can be in two states:
    #   $query is_weak; timer undef => normal user use
    #   $query non-weak; timer exists => due to expire soon
@@ -290,7 +298,8 @@ Takes the following named arguments:
 
 A set of host names are required, either as a named argument or as a
 configured value on the object. If the service name is missing, the default
-CQL port will be used instead.
+CQL port will be used instead. If a host contains a :$service, that overrides
+both the service passed in, and the service on this object.
 
 =cut
 
@@ -301,6 +310,10 @@ sub _connect_node
    my $self = shift;
    my ( $host, $service ) = @_;
 
+   if ($host =~ s/:(\d+)$//){
+      $service = $1;
+      }
+
    $service //= $self->{service} // DEFAULT_CQL_PORT;
 
    my $conn = Net::Async::CassandraCQL::Connection->new(
@@ -308,6 +321,8 @@ sub _connect_node
          my $node = shift;
          $self->remove_child( $node );
          $self->_closed_node( $node->nodeid );
+         # Node close will fail the queries, which must happen after we've chosen a new node
+         $node->on_closed;
       },
       map { $_ => $self->{$_} } qw( username password cql_version ),
    );
@@ -499,7 +514,7 @@ sub _ready_node
       return $conn_f unless keys %$queries_by_cql;
 
       ( fmap_void {
-         my $query = shift->[0] or return Future->new->done;
+         my $query = shift->{query} or return Future->new->done;
          $conn->prepare( $query->cql, $self );
       } foreach => [ values %$queries_by_cql ] )
          ->then( sub { $conn_f } );
@@ -510,28 +525,33 @@ sub _pick_new_eventwatch
 {
    my $self = shift;
 
-   my @primaries = keys %{ $self->{primary_ids} };
+   # Only the primaries which haven't been watched already
+   my @primaries = grep { !$self->{event_ids}{ $_ } } keys %{ $self->{primary_ids} };
 
-   {
-      my $nodeid = $primaries[rand @primaries];
-      redo if $self->{event_ids}{$nodeid};
+   # If there aren't any futues left to try, don't create an infinite loop
+   # Don't expect this to happen, but just in case
+   return Future->fail('No primary node available to eventwatch') unless @primaries;
 
-      $self->{event_ids}{$nodeid} = 1;
+   my $nodeid = $primaries[int rand @primaries];
 
-      my $node = $self->{nodes}{$nodeid};
-      $node->{ready_f}->on_done( sub {
+   $self->{event_ids}{$nodeid} = 1;
+
+   my $node = $self->{nodes}{$nodeid} or die "Expected node: '$nodeid' to exist";
+
+   return $node->{ready_f} = $node->{ready_f}->then( sub {
          my $conn = shift;
          $conn->configure(
             on_topology_change => $self->{on_topology_change_cb},
             on_status_change   => $self->{on_status_change_cb},
          );
          $conn->register( [qw( TOPOLOGY_CHANGE STATUS_CHANGE )] )
-            ->on_fail( sub {
-               delete $self->{event_ids}{$nodeid};
-               $self->_pick_new_eventwatch
-            });
-      });
-   }
+         ->then_done($conn)
+         ->else( sub {
+           delete $self->{event_ids}{$nodeid};
+           # Return the new attempt to eventwatch a differnt primary
+           return $self->_pick_new_eventwatch;
+          });
+    });
 }
 
 sub _on_topology_change
@@ -823,7 +843,8 @@ sub query_rows
 
    $self->query( @_ )->then( sub {
       my ( $type, $result ) = @_;
-      $type eq "rows" or Future->new->fail( "Expected 'rows' result" );
+      defined $type   or return Future->new->fail( "Expected type from query" );
+      $type eq "rows" or return Future->new->fail( "Expected 'rows' result" );
       Future->new->done( $result );
    });
 }
@@ -849,11 +870,11 @@ sub prepare
    my $queries_by_cql = $self->{queries_by_cql};
 
    if( my $q = $queries_by_cql->{$cql} ) {
-      my $query = $q->[0];
-      if( $q->[1] ) {
-         $q->[1]->cancel;
-         undef $q->[1];
-         weaken( $q->[0] );
+      my $query = $q->{query};
+      if( $q->{pacemaker} ) {
+         $q->{pacemaker}->cancel;
+         undef $q->{pacemaker};
+         weaken( $q->{query} );
       }
       return Future->new->done( $query );
    }
@@ -871,13 +892,44 @@ sub prepare
 
       $self->debug_printf( "PREPARE => [%s]", unpack "H*", $query->id );
 
-      my $q = $queries_by_cql->{$cql} = [ $query, undef ];
-      weaken( $q->[0] );
+      my $q = $queries_by_cql->{$cql} = { query => $query };
+      weaken( $q->{query} );
 
       Future->new->done( $query );
    });
 }
 
+# Check if a query cache is the only thing keeping a query alive
+sub _check_query_pacemaker {
+   my ($self, $cql) = @_;
+
+   weaken $self;
+
+   my $q = $self->{queries_by_cql}{$cql};
+
+   my $query = $q->{query};
+
+   $q->{ttl} -= QUERY_PACEMAKER_INTERVAL;
+
+   # 1) Without a loop, there's nothing we can do anyway
+   # 2) Otherwise we've kept this query alive long enough
+   # 3) Otherwise Cassandra and loop is being artificialy kept alive by us.
+   # ( 2 references because Loop has a reference to cassandra in notifiers, and this query has one )
+   if ($q->{ttl} <= 0 || refcount($self) == 2 || !$self->loop){
+
+      # Remove the {cassandra} element from the query so it doesn't
+      # re-register itself for expiry when it is DESTROYed again
+      undef $query->{cassandra};
+
+      delete $self->{queries_by_cql}{$cql};
+   }
+   else {
+      $q->{pacemaker} = $self->loop->delay_future( after => QUERY_PACEMAKER_INTERVAL )->on_done(
+         sub { $self->_check_query_pacemaker( $cql ) }
+      );
+   }
+}
+
 sub _expire_query
 {
    my $self = shift;
@@ -886,15 +938,16 @@ sub _expire_query
    my $queries_by_cql = $self->{queries_by_cql};
    my $q = $queries_by_cql->{$cql} or return;
 
-   my $query = $q->[0]; undef $q->[0]; $q->[0] = $query; # unweaken
+   # Unweaken the query itself, so it'll survive destruction
+   my $query = $q->{query};
+   undef $q->{query};
+   $q->{query} = $query;
+   $q->{ttl}   = QUERY_TTL;
 
-   $q->[1] = $self->loop->delay_future( after => 60 )
-      ->on_done( sub {
-         # Remove the {cassandra} element from the query so it doesn't
-         # re-register itself for expiry when it is DESTROYed again
-         undef $q->[0]{cassandra};
-         delete $queries_by_cql->{$cql};
-      });
+   if ($self->loop){
+     $q->{pacemaker} //= $self->loop->delay_future( after => QUERY_PACEMAKER_INTERVAL )
+        ->on_done( sub { $self->_check_query_pacemaker( $cql ) } );
+   }
 }
 
 =head2 $cass->execute( $query, $data, $consistency, %other_args ) ==> ( $type, $result )
